import numpy as np
from ..hmc import hmc


def lnprob_gaussian(x, icov):
    logp = -np.dot(x, np.dot(icov, x)) / 2.0
    grad = -np.dot(x, icov)
    return logp, grad


def test_1():
    # test sampling from a highly-correlated gaussian
    dim = 2
    x0 = np.zeros(dim)
    cov = np.array([[1, 1.98], [1.98, 4]])
    icov = np.linalg.inv(cov)

    samples, logp, diag = hmc(lnprob_gaussian, x0, args=(icov,),
                  n_samples=10**4, n_burn=10**3,
                  n_steps=10, epsilon=0.20, return_diagnostics=True,
                  return_logp=True, random_state=2)

    C = np.cov(samples, rowvar=0, bias=1)
    np.testing.assert_array_almost_equal(cov, C, 1)
    for i in range(100):
        np.testing.assert_almost_equal(
            lnprob_gaussian(samples[i], icov)[0],
            logp[i])


def test_2():
    # test that random state is used correctly
    dim = 2
    x0 = np.zeros(dim)
    cov = np.array([[1, 1.98], [1.98, 4]])
    icov = np.linalg.inv(cov)

    samples1 = hmc(lnprob_gaussian, x0, args=(icov,),
                  n_samples=10, n_burn=0,
                  n_steps=10, epsilon=0.25, return_diagnostics=False,
                  random_state=0)

    samples2 = hmc(lnprob_gaussian, x0, args=(icov,),
                  n_samples=10, n_burn=0,
                  n_steps=10, epsilon=0.25, return_diagnostics=False,
                  random_state=0)
    np.testing.assert_array_almost_equal(samples1, samples2)



"""
pyhmc: Hamiltonain Monte Carlo Sampling in Python
=================================================

This package is a straight-forward port of the functions `hmc2.m` and
hmc2_opt.m from the [MCMCstuff](http://www.lce.hut.fi/research/mm/mcmcstuff/)
matlab toolbox written by Aki Vehtari. The code is originally based on the
functions `hmc.m` from the [netlab toolbox](http://www.ncrg.aston.ac.uk/netlab/index.php)
written by Ian T Nabney. The portion of algorithm involving "windows" is
derived from the C code for this function included in the
[Software for Flexible Bayesian Modeling](http://www.cs.toronto.edu/~radford/fbm.software.html)
written by Radford Neal.

The original Python [port](https://github.com/koepsell/pyhmc) was made
by Kilian Koepsell, and subsequently modernized by Robert T. McGibbon.

Authors
-------
- Kilian Koepsell <kilian@berkeley.edu>
- Robert T. McGibbon <rmcgibbo@gmail.com>
"""

from __future__ import absolute_import
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
from .hmc import hmc



from __future__ import print_function
import numpy as np
cimport numpy as np
cimport cython
from numpy import zeros, asarray
from numpy cimport npy_intp
from libc.math cimport sqrt, log, exp


@cython.boundscheck(False)
def hmc_main_loop(fun, double[::1] x, args, double[::1] p,
                  double[:, ::1] samples, double[::1] logps,
                  double[:, ::1] diagn_pos, double[:, ::1] diagn_mom,
                  double[::1] diagn_acc, npy_intp opt_nsamples,
                  npy_intp opt_nomit, npy_intp opt_window,
                  npy_intp opt_steps, npy_intp opt_display,
                  npy_intp opt_persistence, npy_intp return_logp,
                  npy_intp return_diagnostics, double alpha, double salpha,
                  double epsilon, random):

    cdef npy_intp n_params = x.shape[0]
    cdef npy_intp n_reject = 0         # number of rejected samples
    cdef npy_intp window_offset = 0    # window offset initialised to zero
    cdef npy_intp k = -opt_nomit       # nomit samples are omitted, so we store
    cdef npy_intp n, stps, direction, have_rej, have_acc
    cdef double a, E, E_old, E_acc, logp, E_rej, H, H_old, acc_free_energy, rej_free_energy

    cdef npy_intp i, j, m
    cdef double[::1] randn, grad
    cdef double[::1] x_old = zeros(n_params)
    cdef double[::1] p_old = zeros(n_params)
    cdef double[::1] x_acc = zeros(n_params)
    cdef double[::1] p_acc = zeros(n_params)
    cdef double[::1] x_rej = zeros(n_params)
    cdef double[::1] p_rej = zeros(n_params)
    cdef double[::1] p_tmp = zeros(n_params)

    # Evaluate starting energy.
    x = x.copy()
    logp, grad = fun(asarray(x), *args)
    E = -logp
    if len(grad) != n_params:
        raise ValueError('fun(x, *args) must return (logp, grad)')

    while k < opt_nsamples:  # samples from k >= 0
        # Store starting position and momenta
        for i in range(n_params):
            x_old[i] = x[i]
        for i in range(n_params):
            p_old[i] = p[i]

        # Recalculate Hamiltonian as momenta have changed
        E_old = E
        # Hold = E + 0.5*(p*p')
        H_old = E
        for i in range(n_params):
            H_old += 0.5 * p[i]*p[i]

        # Decide on window offset, if windowed HMC is used
        if opt_window > 1:
            # window_offset=fix(opt_window*rand(1));
            window_offset = int(opt_window * random.rand())

        have_rej = 0
        have_acc = 0
        n = window_offset
        direction = -1 # the default value for direction
                       # assumes that windowing is used

        while direction == -1 or n != opt_steps:
            # if windowing is not used or we have allready taken
            # window_offset steps backwards...
            if direction == -1 and n == 0:
                # Restore, next state should be original start state.
                if window_offset > 0:
                    for i in range(n_params):
                        x[i] = x_old[i]
                    for i in range(n_params):
                        p[i] = p_old[i]
                    n = window_offset

                # set direction for forward steps
                E = E_old
                H = H_old
                direction = 1
                stps = direction
            else:
                if (n*direction+1 < opt_window) or (n > (opt_steps-opt_window)):
                    # State in the accept and/or reject window.
                    stps = direction
                else:
                    # State not in the accept and/or reject window.
                    stps = opt_steps - 2 * (opt_window - 1)

                # First half-step of leapfrog.
                # p = p - direction*0.5*epsilon.*feval(gradf, x, varargin{:});
                logp, grad = fun(asarray(x), *args)
                for i in range(n_params):
                    p[i] +=  direction * 0.5 * epsilon * grad[i]
                for i in range(n_params):
                    x[i] += direction*epsilon*p[i]

                # Full leapfrog steps.
                # for m = 1:(abs(stps)-1):
                for m in range(int_abs(stps)-1):
                    # p = p - direction*epsilon.*feval(gradf, x, varargin{:});
                    logp, grad = fun(asarray(x), *args)
                    for i in range(n_params):
                        p[i] +=  direction * 0.5 * epsilon * grad[i]
                    for i in range(n_params):
                        x[i] += direction * epsilon * p[i]

                # Final half-step of leapfrog.
                # p = p - direction*0.5*epsilon.*feval(gradf, x, varargin{:});
                logp, grad = fun(asarray(x), *args)
                E = -logp
                for i in range(n_params):
                    p[i] += direction * 0.5 * epsilon * grad[i]

                # H = E + 0.5*(p*p');
                H = E
                for i in range(n_params):
                    H += 0.5 * p[i] * p[i]

                n += stps

            if opt_window != opt_steps+1 and n < opt_window:
                # Account for state in reject window.  Reject window can be
                # ignored if windows consist of the entire trajectory.
                if not have_rej:
                    rej_free_energy = H
                else:
                    rej_free_energy = -addlogs(-rej_free_energy, -H)

                if not have_rej or random.rand() < exp(rej_free_energy-H):
                    E_rej = E
                    for i in range(n_params):
                        x_rej[i] = x[i]
                    for i in range(n_params):
                        p_rej[i] = p[i]
                    have_rej = 1

            if n > (opt_steps-opt_window):
                # Account for state in the accept window.
                if not have_acc:
                    acc_free_energy = H
                else:
                    acc_free_energy = -addlogs(-acc_free_energy, -H)

                if not have_acc or random.rand() < exp(acc_free_energy-H):
                    E_acc = E
                    for i in range(n_params):
                        x_acc[i] = x[i]
                    for i in range(n_params):
                        p_acc[i] = p[i]
                    have_acc = 1

        # Acceptance threshold.
        a = exp(rej_free_energy - acc_free_energy)

        if return_diagnostics and k >= 0:
            j = k
            for i in range(n_params):
                diagn_pos[j,i] = x_acc[i]
                diagn_mom[j,i] = p_acc[i]
            diagn_acc[j] = a

        if opt_display:
            print('New position is\n', np.asarray(x))

        # Take new state from the appropriate window.
        if a > random.rand():
            # Accept
            E = E_acc
            for i in range(n_params):
                x[i] = x_acc[i]
            for i in range(n_params):
                p[i] = -p_acc[i]  # Reverse momenta
            if opt_display:
                print('Finished step %4d  Threshold: %g\n'%(k,a))
        else:
            # Reject
            if k >= 0:
                n_reject = n_reject + 1

            E = E_rej
            for i in range(n_params):
                x[i] = x_rej[i]
            for i in range(n_params):
                p[i] = p_rej[i]
            if opt_display:
                print('  Sample rejected %4d.  Threshold: %g\n'%(k,a))

        if k >= 0:
            j = k
            # Store sample
            for i in range(n_params):
                samples[j, i] = x[i]
            if return_logp:
                # Store energy
                logps[j] = -E

        # Set momenta for next iteration
        if opt_persistence:
            # Reverse momenta
            for i in range(n_params):
                p[i] = -p[i]
            # Adjust momenta by a small random amount
            randn = random.randn(n_params)
            for i in range(n_params):
                p[i] = alpha*p[i] + salpha*randn[i]
        else:
            # Replace all momenta
            randn = random.randn(n_params)
            for i in range(n_params):
                p[i] = randn[i]

        k += 1

    return n_reject


cdef inline double addlogs(double a, double b):
    if a > b:
        return a + log(1 + exp(b-a))
    else:
        return b + log(1 + exp(a-b))


cdef inline int int_abs(int a):
    return a if a > 0 else -a




"""Hybrid Monte Carlo for Python

This package is a straight-forward port of the functions ``hmc2.m`` and
``hmc2_opt.m`` from the
`MCMCstuff <http://www.lce.hut.fi/research/mm/mcmcstuff/>`__ matlab
toolbox written by Aki Vehtari. The code is originally based on the
functions ``hmc.m`` from the `netlab
toolbox <http://www.ncrg.aston.ac.uk/netlab/index.php>`__ written by Ian
T Nabney. The portion of algorithm involving "windows" is derived from
the C code for this function included in the `Software for Flexible
Bayesian
Modeling <http://www.cs.toronto.edu/~radford/fbm.software.html>`__
written by Radford Neal.

The original Python `port <https://github.com/koepsell/pyhmc>`__ was
made by Kilian Koepsell, and subsequently modernized by Robert T.
McGibbon.

Authors
-------

-  Kilian Koepsell kilian@berkeley.edu
-  Robert T. McGibbon rmcgibbo@gmail.com

This software is distributed under the BSD License (see LICENSE file).
"""
from setuptools import find_packages, setup, Extension
import numpy as np
from Cython.Distutils import build_ext
import versioneer
versioneer.VCS = 'git'
versioneer.versionfile_source = 'pyhmc/_version.py'
versioneer.versionfile_build = 'pyhmc/_version.py'
versioneer.tag_prefix = '' # tags are like 1.2.0
versioneer.parentdir_prefix = 'pyhmc-' # dirname like 'myproject-1.2.0'

DOCLINES = __doc__.split("\n")
CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: BSD
Programming Language :: Python
Operating System :: OS Independent
"""

cmdclass = versioneer.get_cmdclass()
cmdclass['build_ext'] = build_ext

setup(
    name='pyhmc',
    author="Robert T. McGibbon",
    author_email='rmcgibbo@gmail.com',
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    url="https://github.com/rmcgibbo/pyhmc",
    description=DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    license='BSD',
    install_requires=['numpy'],
    packages=find_packages(),
    zip_safe=False,
    ext_modules=[Extension('pyhmc._hmc', ['pyhmc/_hmc.pyx'],
                           include_dirs=[np.get_include()])],
)



from __future__ import print_function, division, absolute_import

import numbers
import numpy as np
from ._hmc import hmc_main_loop

__all__ = ['hmc', '__version__']


def hmc(fun, x0, n_samples=1000, args=(), display=False, n_steps=1, n_burn=0,
        persistence=False, decay=0.9, epsilon=0.2, window=1,
        return_logp=False, return_diagnostics=False, random_state=None):
    """Hamiltonian Monte Carlo sampler.

    Uses a Hamiltonian / Hybrid Monte Carlo algorithm to sample from the
    distribution P ~ exp(f). The Markov chain starts at the point x0. The
    callable ``fun`` should return the log probability and gradient of the
    log probability of the target density.

    Parameters
    ----------
    fun : callable
        A callable which takes a vector in the parameter spaces as input
        and returns the natural logarithm of the posterior probabily
        for that position, and gradient of the posterior probability with
        respect to the parameter vector, ``logp, grad = func(x, *args)``.
    x0 : 1-d array
      Starting point for the sampling Markov chain.

    Other Parameters
    ----------------
    n_samples : int
        The number of samples retained from the Markov chain.
    args : tuple
        additional arguments to be passed to fun().
    display : bool
        If True, enables verbose display output. Default: False
    n_steps : int
        Defines the trajectory length between Metropolized steps (i.e. the
        number of leapfrog steps at each iteration before accept/reject).
    n_burn : int
        The number of samples omitted from the start of the chain as 'burn in'.
    persistence : bool
        If True, momentum persistence is used (i.e. the momentum
        variables decay rather than being replaced). Default: False
    decay : float, default=0.9
        Defines the decay used when a persistent update of (leap-frog)
        momentum is used. Bounded to the interval [0, 1.).
    epsilon : float, default=0.2.
        The step adjustment used in the integrator. In physics-terms, this
        is the time step.
    window : int, default=1
        The size of the acceptance window (See [1], Section 5.4)
    return_logp : bool, default=False
        If True, the energy values for all samples are returned.
    return_diagnostics : bool, default=False
        If True, diagnostic information is returned (see below).

    Returns
    -------
    samples : array, shape=(n_samples, n_params)
        Array with data samples in rows.
    logp : array, shape=(n_samples)
        If return_logp is True, also returns an array of the log probability
        for all samples.
    diagn : dict
        If return_diagnostics is True, also returns a dictionary with diagnostic
        information (position, momentum and acceptance threshold) for each step
        of the chain in diagn.pos, diagn.mom and diagn.acc respectively.
        All candidate states (including rejected ones) are stored in
        diagn['pos']. The diagn dictionary contains the following items:

          ``pos`` : array
             the position vectors of the dynamic process
          ``mom`` : array
             the momentum vectors of the dynamic process
          ``acc`` : array
             the acceptance thresholds
          ``rej`` : float
             the rejection rate
          ``stp`` : float
             the step size vectors

    References
    ----------
    .. [1] Neal, Radford. "MCMC using Hamiltonian dynamics." Handbook of
       Markov Chain Monte Carlo 2 (2011).
    """
    # check some options
    assert n_steps >= 1, 'step size has to be >= 1'
    assert n_samples >= 1, 'n_samples has to be >= 1'
    assert n_burn >= 0, 'n_burn has to be >= 0'
    assert decay >= 0, 'decay has to be >= 0'
    assert decay <= 1, 'decay has to be <= 1'
    assert window >= 0, 'window has to be >= 0'
    if window > n_steps:
        window = n_steps
        if display:
            print("setting window size to step size %d" % window)

    if persistence:
        alpha = decay
        salpha = np.sqrt(1-alpha**2)
    else:
        alpha = salpha = 0.

    n_params = len(x0)

    # Initialize matrix of returned samples
    samples = np.zeros((n_samples, n_params))

    # Return energies?
    if return_logp:
        logp = np.zeros(n_samples)
    else:
        logp = np.zeros(0)

    # Return diagnostics?
    if return_diagnostics:
        diagn_pos = np.zeros((n_samples, n_params))
        diagn_mom = np.zeros((n_samples, n_params))
        diagn_acc = np.zeros(n_samples)
    else:
        diagn_pos = None
        diagn_mom = None
        diagn_acc = None

    random = _check_random_state(random_state)
    p = random.randn(n_params)

    # Main loop.
    all_args = [
        fun, x0, args, p, samples, logp,
        diagn_pos, diagn_mom, diagn_acc,
        n_samples, n_burn, window,
        n_steps, display, persistence,
        return_logp, return_diagnostics,
        alpha, salpha, epsilon, random,]

    n_reject = hmc_main_loop(*all_args)

    if display:
        print('\nFraction of samples rejected:  %g\n'%(n_reject / n_samples))

    # Store diagnostics
    if return_diagnostics:
        diagn = dict()
        diagn['pos'] = diagn_pos   # positions matrix
        diagn['mom'] = diagn_mom   # momentum matrix
        diagn['acc'] = diagn_acc   # acceptance treshold matrix
        diagn['rej'] = n_reject / n_samples   # rejection rate
        diagn['stps'] = epsilon    # stepsize vector

    if return_logp or return_diagnostics:
        out = (samples,)
    else:
        return samples

    if return_logp:
        out += (logp,)
    if return_diagnostics:
        out += (diagn,)
    return out



def _check_random_state(seed):
    """Turn seed into a np.random.RandomState instance

    If seed is None, return the RandomState singleton used by np.random.
    If seed is an int, return a new RandomState instance seeded with seed.
    If seed is already a RandomState instance, return it.
    Otherwise raise ValueError.
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)
